#include "oslp.h"

extern Handle *handle;

/* GNode Traversal Callbacks */
static gboolean TraverseFunc(GNode * node, gpointer data);
static gboolean FindFunc(GNode * node, gpointer d);
static gboolean DestroyFunc(GNode * node, gpointer data);


/* Parse the  SLP service URL and return a name to be displayed.*/
char *return_display_name(char *slp_service_uri)
{
	g_return_val_if_fail(slp_service_uri != NULL, NULL);

	SLPError err;
	SLPSrvURL *srv_url;
	char *reqstr = NULL;

	err = SLPParseSrvURL(slp_service_uri, &srv_url);

	if (strcmp(srv_url->s_pcHost, "") == 0)
		if (*srv_url->s_pcSrvPart == '/')
			reqstr =
			    g_strdup_printf("%s",
					    srv_url->s_pcSrvPart + 1);
		else
			reqstr =
			    g_strdup_printf("%s", srv_url->s_pcSrvPart);
	else
		reqstr = g_strdup_printf("%s", srv_url->s_pcHost);

	return (reqstr);
}

/* Take a GnomeVFSURI and return if it is a directory or not. */
gboolean is_directory(GnomeVFSURI * uri)
{
	g_return_val_if_fail(uri != NULL, -1);

	int uri_type;

	uri_type = determine_uri_type(uri);
	if (uri_type == SLP_SERVER_NAME)
		return FALSE;
	else
		return TRUE;
}


static gboolean FindFunc(GNode * node, gpointer search_data)
{
	SlpHandle *cur_handle;

	cur_handle = (SlpHandle *) node->data;
	if (strcmp(cur_handle->display_data, search_data) == 0) {
		handle->find_child = node;
		return TRUE;
	}

	return FALSE;
}


GNode *my_g_node_find(GNode * root, const char *search_data)
{
	g_assert(root);
	g_assert(search_data);

	g_node_traverse(root,
			G_PRE_ORDER,
			G_TRAVERSE_ALL, -1, FindFunc,
			(gpointer) search_data);

	return handle->find_child;
}


SlpHandle *give_tmp_slp_handle(const char *display_data, gboolean is_dir)
{
	g_return_val_if_fail(display_data != NULL, NULL);

	SlpHandle *tmp = NULL;
	tmp = g_new0(SlpHandle, 1);
	tmp->is_dir = is_dir;
	tmp->display_data = g_strdup(display_data);
	tmp->data = NULL;
	return tmp;
}

/* Extract the appropriate part of the GnomeVFSURI like "slp:///" 
or "ldap.novell" or "164.99.152.151:389", depending on the type passed. */

char *parse_and_extract(GnomeVFSURI * uri, int type)
{

	g_return_val_if_fail(uri != NULL, NULL);
	g_return_val_if_fail(type <= 2 && type >= 0, NULL);

	char *text_uri;
	char **token_array1, **token_array2;
	char *ret_str = NULL;

	text_uri = gnome_vfs_uri_to_string(uri, GNOME_VFS_URI_HIDE_NONE);

	/* Tokenise on ":///" and further tokenise the second token on "/" */
	token_array1 = g_strsplit(text_uri, ":///", 10);
	token_array2 = g_strsplit(token_array1[1], "/", 10);

	if (type == SLP_SERVICES) {
		ret_str = g_strdup(token_array1[0]);
	}

	else if (type == SLP_SERVICE_TYPE) {
		ret_str = g_strdup(token_array2[0]);
	} else if (type == SLP_SERVER_NAME) {
		ret_str = g_strdup(token_array2[1]);
	}

	g_strfreev(token_array2);
	g_strfreev(token_array1);
	g_free(text_uri);

	return ret_str;
}

/* Take a GnomeVFSURI and return what "uri_type" it is. */
int determine_uri_type(GnomeVFSURI * uri)
{
	g_return_val_if_fail(uri != NULL, -1);

	char *text_uri;
	char **token_array1, **token_array2;

	int i = 0;
	int j = 0;

	text_uri = gnome_vfs_uri_to_string(uri, GNOME_VFS_URI_HIDE_NONE);

	token_array1 = g_strsplit(text_uri, ":///", 10);
	while (token_array1[i] != NULL)
		i = i + 1;

	if (i == 2 && strcmp(token_array1[i - 1], "") == 0) {
		g_strfreev(token_array1);
		g_free(text_uri);
		return SLP_SERVICES;
	} else if (i > 2) {
		g_strfreev(token_array1);
		g_free(text_uri);
		return -1;
	}

	token_array2 = g_strsplit(token_array1[i - 1], "/", 10);
	while (token_array2[j] != NULL)
		j = j + 1;

	g_strfreev(token_array2);
	g_strfreev(token_array1);
	g_free(text_uri);

	if (j == 1)
		return SLP_SERVICE_TYPE;
	else if (j == 2)
		return SLP_SERVER_NAME;
	else
		return -1;
}

static gboolean TraverseFunc(GNode * node, gpointer d)
{

	if (node) {
		SlpHandle *cur_handle;
		cur_handle = (SlpHandle *) node->data;
		printf(" display_data:%s \n  data:%s\n",
		       cur_handle->display_data, cur_handle->data);
	}

	return FALSE;
}

/* Traverse the entire tree for displating its contents. */
void parse_tree(GNode * root)
{

	g_node_traverse(root,
			G_PRE_ORDER, G_TRAVERSE_ALL, -1, TraverseFunc,
			NULL);
}

/* Free the dynamically allocated members of GNode->data ie. SlpHandles.
Not needed now, g_node_destroy () frees them up automatically ?????? */

static gboolean DestroyFunc(GNode * node, gpointer d)
{

	if (node) {
		SlpHandle *cur_handle;
		cur_handle = (SlpHandle *) node->data;
		g_free(cur_handle->data);
		g_free(cur_handle->display_data);
	}

	return FALSE;
}


void my_g_node_destroy(GNode * root)
{

	g_node_traverse(root,
			G_PRE_ORDER, G_TRAVERSE_ALL, -1, DestroyFunc,
			NULL);
}
