#include "oslp.h"

Handle *handle = NULL;
static gboolean handle_initialised = FALSE;

G_LOCK_DEFINE_STATIC(handle);

/* SLP Callback for locating attributes of a Service URL */
static SLPBoolean FindAttrsCB(SLPHandle openslp_handle,
			      const char *attrlist,
			      SLPError errcode, void *cookie);

/* SLP Callback for discovering Service Types */
static SLPBoolean FindSrvTypesCB(SLPHandle openslp_handle,
				 const char *srvtypes,
				 SLPError errcode, void *cookie);

/* SLP Callback for discovering servers providing a particular service */
static SLPBoolean FindSrvsCB(SLPHandle openslp_handle,
			     const char *srvurl,
			     unsigned short lifetime,
			     SLPError errcode, void *cookie);


/* Initialise the Global Handle */
static GnomeVFSResult init_handle()
{
	SlpHandle *root;

	if (handle_initialised == TRUE)
		return GNOME_VFS_OK;

	handle = g_new0(Handle, 1);
	root = give_tmp_slp_handle("Root", TRUE);
	handle->root = g_node_new((gpointer) root);
	handle->cur_child = NULL;
	handle->find_child = NULL;

	handle_initialised = TRUE;
	return GNOME_VFS_OK;
}

/* Deinitialise the global Handle */
static void deinit_handle()
{
	if (handle_initialised == FALSE)
		return;
	/* We assume g_node_destroy () frees all the SlpHandles too. 
	   Explicit attempts to free the structure members lead to crashes. */

	//      my_g_node_destroy (handle->root);
	g_node_destroy(handle->root);
	handle->root = NULL;

	g_free(handle);
	handle_initialised = FALSE;
}


static SLPBoolean FindSrvTypesCB(SLPHandle openslp_handle,
				 const char *srvtypes,
				 SLPError errcode, void *cookie)
{

	if (errcode == SLP_OK && *srvtypes) {
		char *token;
		char *buff, rbuff[200];

		buff = strdup(srvtypes);

		token = strtok_r(buff, ",", (char **) &rbuff);
		while (token) {
			SlpHandle *cur_handle;

			cur_handle = give_tmp_slp_handle(token, TRUE);
			/* Add the SlpHandles with the Service Type Tokens 
				as children of the tree root. */

			handle->cur_child =
			    g_node_append_data(handle->root, cur_handle);
			token = strtok_r(NULL, ",", (char **) &rbuff);
		}

		g_free(buff);

		return SLP_FALSE;
	}

	else if (errcode == SLP_LAST_CALL) {
		return SLP_FALSE;
	}

	return SLP_TRUE;
}



SLPBoolean FindAttrsCB(SLPHandle openslp_handle,
		       const char *attrlist,
		       SLPError errcode, void *cookie)
{
	char *buff = NULL;

	if (errcode == SLP_OK && *attrlist) {

		GNode *server = (GNode *) cookie;
		SlpHandle *cur_handle;

		buff = g_strdup(attrlist);
		g_strdelimit(buff, "(", '\r');
		g_strdelimit(buff, ")", '\n');
		g_strdelimit(buff, ",", '\n');
		strcat(buff, "\n\n");

		cur_handle = give_tmp_slp_handle(buff, FALSE);

		/* Add all the attributes as SlpHandle struct members to 
			the ServerName passed in as cookie to FindAttrsCB */

		handle->cur_child = g_node_append_data(server, cur_handle);
		g_free(buff);

		return SLP_FALSE;
	}

	else if (errcode == SLP_LAST_CALL) {
		return SLP_FALSE;
	}

	return SLP_TRUE;

}

static SLPBoolean FindSrvsCB(SLPHandle openslp_handle,
			     const char *srvurl,
			     unsigned short lifetime,
			     SLPError errcode, void *cookie)
{

	g_assert(cookie);

	if (errcode == SLP_OK && *srvurl) {
		char *token;
		char *buff, rbuff[200];

		buff = strdup(srvurl);
		token = strtok_r(buff, ",", (char **) rbuff);
		while (token != NULL) {
			SlpHandle *cur_handle;
			char *display_name;

			display_name = return_display_name(token);
			cur_handle =
			    give_tmp_slp_handle(display_name, FALSE);
			cur_handle->data = g_strdup(token);

			/* Add the server names as SlpHandles to the 
			ServiceType node passed in as cookie to this callback */
			if (cookie)
				handle->cur_child =
				    g_node_append_data((GNode *) cookie,
						       (SlpHandle *)
						       cur_handle);

			g_free(display_name);
			token = strtok_r(NULL, ",", (char **) rbuff);
		}

		g_free(buff);

		/* Gross Hack : 
		   Despite opening a synchronous SLP handle, 
		   this callback gives me *one* service URL at a time 
		   sometimes. Returning SLP_TRUE to get all the URLs */

		return SLP_TRUE;
	}

	else if (errcode == SLP_LAST_CALL) {
		return SLP_FALSE;
	}

	return SLP_FALSE;
}



static GnomeVFSResult slp_handle_srv_types()
{
	SLPError result;

	result = SLPOpen(NULL,
			 SLP_FALSE,
			 (SLPHandle *) & handle->openslp_handle);
	if (result != SLP_OK) {
		return GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
	}


	result =
	    SLPFindSrvTypes(handle->openslp_handle, "*", "",
			    FindSrvTypesCB, NULL);

	if (result != SLP_OK) {
		SLPClose(handle->openslp_handle);
		return GNOME_VFS_ERROR_NOT_FOUND;
	}

	SLPClose(handle->openslp_handle);
	return GNOME_VFS_OK;
}



static GnomeVFSResult slp_handle_servers(GnomeVFSURI * uri)
{

	g_assert(uri);

	SLPError result;
	GNode *srv_type_node = NULL;
	char *srv_type;

	result =
	    SLPOpen(NULL, SLP_FALSE,
		    (SLPHandle *) & handle->openslp_handle);
	if (result != SLP_OK) {
		return GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
	}

	/* Find under which ServiceType node to insert the Slp Service URLs. */
	srv_type = parse_and_extract(uri, SLP_SERVICE_TYPE);
	srv_type_node = my_g_node_find(handle->root, srv_type);

	if (srv_type_node == NULL) {
		g_free(srv_type);
		SLPClose(handle->openslp_handle);
		return GNOME_VFS_ERROR_NOT_FOUND;
	}


	result = SLPFindSrvs(handle->openslp_handle, srv_type, "", "",
			     FindSrvsCB, (void *) srv_type_node);

	g_free(srv_type);
	if (result != SLP_OK) {
		SLPClose(handle->openslp_handle);
		return GNOME_VFS_ERROR_NOT_FOUND;
	}

	SLPClose(handle->openslp_handle);
	return GNOME_VFS_OK;
}


static GnomeVFSResult
slp_open(GnomeVFSMethod * method,
	 GnomeVFSMethodHandle ** method_handle,
	 GnomeVFSURI * uri,
	 GnomeVFSOpenMode mode, GnomeVFSContext * context)
{

	char *server_name = NULL;
	SLPError result;
	int uri_type;
	GNode *server = NULL;
	SlpHandle *cur_handle;
	GnomeVFSResult res;

	if (mode & GNOME_VFS_OPEN_RANDOM)
		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;

	if (mode & GNOME_VFS_OPEN_WRITE)
		return GNOME_VFS_ERROR_READ_ONLY;

	uri_type = determine_uri_type(uri);
	if (uri_type != SLP_SERVER_NAME)
		return GNOME_VFS_ERROR_INVALID_URI;

	G_LOCK(handle);
	/* Build the tree again. */

	/* TODO: This is a performance hit for bigger networks. 
	Figure out a way to handle this without building the tree again.*/

	init_handle();
	res = slp_handle_srv_types();
	if (res != GNOME_VFS_OK) {
		deinit_handle();
		G_UNLOCK(handle);
		return res;
	}

	res = slp_handle_servers(uri);
	if (res != GNOME_VFS_OK) {
		deinit_handle();
		G_UNLOCK(handle);
		return res;
	}

	/* Find under what server name/SLP Service URL to insert the 
	service attributes. */

	server_name = parse_and_extract(uri, SLP_SERVER_NAME);
	server = my_g_node_find(handle->root, server_name);
	g_free(server_name);

	if (server == NULL) {
		deinit_handle();
		G_UNLOCK(handle);
		return GNOME_VFS_ERROR_HOST_NOT_FOUND;

	}

	cur_handle = (SlpHandle *) server->data;

	result =
	    SLPOpen(NULL, SLP_FALSE,
		    (SLPHandle *) & handle->openslp_handle);

	if (result != SLP_OK) {
		deinit_handle();
		G_UNLOCK(handle);
		return GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
	}

	result =
	    SLPFindAttrs(handle->openslp_handle, cur_handle->data, "", "",
			 FindAttrsCB, server);

	if (result != SLP_OK) {
		SLPClose(handle->openslp_handle);
		deinit_handle();
		G_UNLOCK(handle);
		return GNOME_VFS_ERROR_INTERNAL;
	}

	SLPClose(handle->openslp_handle);
	G_UNLOCK(handle);
	return GNOME_VFS_OK;
}


static GnomeVFSResult
slp_close(GnomeVFSMethod * method,
	  GnomeVFSMethodHandle * method_handle, GnomeVFSContext * context)
{

	G_LOCK(handle);
	//      parse_tree(handle->root);
	deinit_handle();
	G_UNLOCK(handle);

	return GNOME_VFS_OK;
}

static GnomeVFSResult
slp_read(GnomeVFSMethod * method,
	 GnomeVFSMethodHandle * method_handle,
	 gpointer buffer,
	 GnomeVFSFileSize num_bytes,
	 GnomeVFSFileSize * bytes_read, GnomeVFSContext * context)
{

	static GnomeVFSFileSize bytes_written;
	GnomeVFSFileSize cur_data_size = 0;
	SlpHandle *cur_handle = NULL;

	G_LOCK(handle);
	if (handle->cur_child == NULL) {
		G_UNLOCK(handle);
		*bytes_read = 0;
		return GNOME_VFS_ERROR_EOF;
	}

	cur_handle = (SlpHandle *) ((handle->cur_child)->data);
	if (cur_handle) {
		cur_data_size = strlen(cur_handle->display_data) + 1;
		*bytes_read = MIN(num_bytes, cur_data_size);
		memcpy(buffer, cur_handle->display_data, *bytes_read);
		bytes_written += cur_data_size;

		//      handle->cur_child = handle->cur_child->prev;
		/* We add all the attributes to display_data member,
		 no need to set the cur_child */

	} else {
		G_UNLOCK(handle);
		return GNOME_VFS_ERROR_CORRUPTED_DATA;
	}

	G_UNLOCK(handle);
	return GNOME_VFS_ERROR_EOF;
}


static GnomeVFSResult
slp_open_directory(GnomeVFSMethod * method,
		   GnomeVFSMethodHandle ** method_handle,
		   GnomeVFSURI * uri,
		   GnomeVFSFileInfoOptions options,
		   GnomeVFSContext * context)
{
	int type = -1;

	type = determine_uri_type(uri);

	/* Handle the "slp:///" URI. */
	if (type == SLP_SERVICES) {
		GnomeVFSResult res;

		G_LOCK(handle);
		init_handle();
		res = slp_handle_srv_types();
		if (res != GNOME_VFS_OK)
			deinit_handle();

		G_UNLOCK(handle);
		return res;
	}

	/* Handle the slp:///ldap.novell URIs ie. the Service Type URIs. */
	else if (type == SLP_SERVICE_TYPE) {
		GnomeVFSResult res;

		G_LOCK(handle);
		init_handle();

		/* Build the tree again, See comment above in slp_open (). */
		res = slp_handle_srv_types();
		if (res != GNOME_VFS_OK) {
			deinit_handle();
			G_UNLOCK(handle);
			return res;
		}

		res = slp_handle_servers(uri);
		if (res != GNOME_VFS_OK) {
			deinit_handle();
			G_UNLOCK(handle);
			return res;
		}

		G_UNLOCK(handle);
		return GNOME_VFS_OK;

	} else {
		return GNOME_VFS_ERROR_INVALID_URI;
	}

}


static GnomeVFSResult
slp_close_directory(GnomeVFSMethod * method,
		    GnomeVFSMethodHandle * method_handle,
		    GnomeVFSContext * context)
{

	G_LOCK(handle);
	//              parse_tree(handle->root);
	deinit_handle();
	G_UNLOCK(handle);

	return GNOME_VFS_OK;
}

static GnomeVFSResult
slp_read_directory(GnomeVFSMethod * method,
		   GnomeVFSMethodHandle * method_handle,
		   GnomeVFSFileInfo * file_info, GnomeVFSContext * context)
{

	SlpHandle *cur_child_data;

	G_LOCK(handle);

	if (!handle->cur_child) {
		G_UNLOCK(handle);
		return GNOME_VFS_ERROR_EOF;
	}

	cur_child_data = (SlpHandle *) handle->cur_child->data;

	if (cur_child_data->is_dir) {
		file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
		file_info->mime_type = g_strdup("x-directory/normal");
		file_info->valid_fields |=
		    GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
	} else {
		file_info->type = GNOME_VFS_FILE_TYPE_REGULAR;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
		file_info->mime_type = g_strdup("text/plain");
		file_info->valid_fields |=
		    GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
		/* TODO: Set the filesize properly. */
		file_info->size = 500;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_SIZE;
	}

	file_info->name = cur_child_data->display_data;

	/* cur_child points to the last child after g_node_append_data ().
	 Set it to point to the previous sibling. */

	handle->cur_child = handle->cur_child->prev;

	G_UNLOCK(handle);

	return GNOME_VFS_OK;
}

static GnomeVFSResult
slp_get_file_info(GnomeVFSMethod * method,
		  GnomeVFSURI * uri,
		  GnomeVFSFileInfo * file_info,
		  GnomeVFSFileInfoOptions options,
		  GnomeVFSContext * context)
{
	gboolean is_dir;
	int uri_type = -1;
	char *name = NULL;

	/* Set the name to the relevant part of the URI passed. */
	uri_type = determine_uri_type(uri);
	name = parse_and_extract(uri, uri_type);
	file_info->name = g_strdup(name);

	is_dir = is_directory(uri);
	if (is_dir) {
		file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
		file_info->mime_type = g_strdup("x-directory/normal");
		file_info->valid_fields |=
		    GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
	} else {
		file_info->type = GNOME_VFS_FILE_TYPE_REGULAR;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
		file_info->mime_type = g_strdup("text/plain");
		file_info->valid_fields |=
		    GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
		//      file_info->size = 500;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_SIZE;
	}

	g_free(name);
	return GNOME_VFS_OK;
}

static gboolean
slp_is_local(GnomeVFSMethod * method, const GnomeVFSURI * uri)
{
	return FALSE;
}

/* We only support reading, opening and closing of files and directories */
static GnomeVFSMethod method = {
	sizeof(GnomeVFSMethod),
	slp_open,
	NULL,			//        slp_create
	slp_close,
	slp_read,
	NULL,			//        slp_write
	NULL,			//        slp_seek
	NULL,			//        slp_tell
	NULL,			//        slp_truncate_handle
	slp_open_directory,
	slp_close_directory,
	slp_read_directory,
	slp_get_file_info,
	NULL,			//        slp_get_file_info_from_handle
	slp_is_local,
	NULL,			//        slp_make_directory
	NULL,			//        slp_remove_directory
	NULL,			//        slp_move
	NULL,			//        slp_unlink
	NULL,			//        slp_check_same_fs
	NULL,			//        slp_set_file_info
	NULL,			//        slp_truncate
	NULL,			//        slp_find_directory
	NULL			//        slp_create_symbolic_link
};


GnomeVFSMethod *vfs_module_init(const char *method_name, const char *args)
{

	if ((strncmp(method_name, SLP_INITIAL_URI, SLP_INITIAL_URI_LENGTH))
	    == 0) {
		GnomeVFSResult result;

		G_LOCK(handle);
		result = init_handle();
		G_UNLOCK(handle);

		if (result == GNOME_VFS_OK)
			return &method;
	}

	return NULL;
}



void vfs_module_shutdown(GnomeVFSMethod * method)
{
	G_LOCK(handle);
	deinit_handle();
	G_UNLOCK(handle);
}
